# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ArithmeticRightShiftAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ArithmeticRightShiftAttribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsArithmeticRightShiftAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def ArithmeticRightShiftAttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x4F\x53\x41", size_prefixed=size_prefixed)

    # ArithmeticRightShiftAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ArithmeticRightShiftAttribute
    def Round(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def ArithmeticRightShiftAttributeStart(builder):
    builder.StartObject(1)

def Start(builder):
    ArithmeticRightShiftAttributeStart(builder)

def ArithmeticRightShiftAttributeAddRound(builder, round):
    builder.PrependBoolSlot(0, round, 0)

def AddRound(builder, round):
    ArithmeticRightShiftAttributeAddRound(builder, round)

def ArithmeticRightShiftAttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return ArithmeticRightShiftAttributeEnd(builder)
