//
// SPDX-FileCopyrightText: Copyright 2023-2025 Arm Limited and/or its affiliates <open-source-office@arm.com>
//
// SPDX-License-Identifier: Apache-2.0
//
// Licensed under the Apache License, Version 2.0 (the License); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an AS IS BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Automatically generated by tosaValidationGenerator for TOSA Specification 1.0.0draft
// Do not edit.

#include "tosa/tosa_validator.hpp"

#include "compiler/operation.hpp"

namespace tosa
{
namespace validator
{

void ValidateOperator(const GraphApi::GraphOperation *graphOp, const Context &context)
{
    if ( graphOp == nullptr ) throw std::invalid_argument("No operation");
    if ( (context.version & 0xFFFFFF00) == GraphApi::VERSION_TOSA_1_00 && context.profile == GraphApi::PROFILE_BASELINE )
    {
        ValidateOperator_Version_1_0_0_draft_Profile_PRO_INT(graphOp, context);
        return;
    }
    throw std::invalid_argument("TOSA version or profile not supported");
}

}  // namespace validator
}  // namespace tosa
