//
// SPDX-FileCopyrightText: Copyright 2023-2025 Arm Limited and/or its affiliates <open-source-office@arm.com>
//
// SPDX-License-Identifier: Apache-2.0
//
// Licensed under the Apache License, Version 2.0 (the License); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an AS IS BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Automatically generated by tosaValidationGenerator for TOSA Specification 1.0.0draft
// Do not edit.

#pragma once

#include "tosa/tosa_validator.hpp"

namespace tosa
{
namespace validator
{
namespace checks
{
// Checks for TOSA Specification 1.0.0draft
void RequireCheck_3gogyrefl20gp(const regor::Operation *op, const Context &context);
void RequireCheck_35z4hcgn21c8p(const regor::Operation *op, const Context &context);
void RequireCheck_2f51h19mqfhr8(const regor::Operation *op, const Context &context);
void RequireCheck_3jqx5d6a2c85r(const regor::Operation *op, const Context &context);
void RequireCheck_1b64l72fvni7o(const regor::Operation *op, const Context &context);
void RequireCheck_3otz8rylb4eh1(const regor::Operation *op, const Context &context);
void RequireCheck_2p74g4god707n(const regor::Operation *op, const Context &context);
void RequireCheck_61j2lms4vo0v(const regor::Operation *op, const Context &context);
void RequireCheck_3nkub9jwwaf4h(const regor::Operation *op, const Context &context);
void RequireCheck_3o6eotvyt76cz(const regor::Operation *op, const Context &context);
void RequireCheck_31n0oq4yculbk(const regor::Operation *op, const Context &context);
void RequireCheck_2apk8ly9uthz6(const regor::Operation *op, const Context &context);
void RequireCheck_3ah9e1mnk126p(const regor::Operation *op, const Context &context);
void RequireCheck_2e13btqfr98am(const regor::Operation *op, const Context &context);
void RequireCheck_lt5trq6bbw9w(const regor::Operation *op, const Context &context);
void RequireCheck_1nuiu459z8num(const regor::Operation *op, const Context &context);
void RequireCheck_182ljbxwn59zs(const regor::Operation *op, const Context &context);
void RequireCheck_2pd0619ns6vtd(const regor::Operation *op, const Context &context);
void RequireCheck_12cwj1gynxopa(const regor::Operation *op, const Context &context);
void RequireCheck_7uvvy4pqp2pj(const regor::Operation *op, const Context &context);
void RequireCheck_2wyo0jz6whe2p(const regor::Operation *op, const Context &context);
void RequireCheck_2btzv7wkv70o0(const regor::Operation *op, const Context &context);
}  // namespace checks
}  // namespace validator
}  // namespace tosa
