//
// SPDX-FileCopyrightText: Copyright 2023-2025 Arm Limited and/or its affiliates <open-source-office@arm.com>
//
// SPDX-License-Identifier: Apache-2.0
//
// Licensed under the Apache License, Version 2.0 (the License); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an AS IS BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Automatically generated by tosaValidationGenerator for TOSA Specification 1.0.0draft
// Do not edit.

#pragma once

#include "tosa/tosa_validator.hpp"

namespace tosa
{
namespace validator
{
namespace checks
{
// Checks for TOSA Specification 1.0.0draft
void ErrorIfCheck_3tg4p2a5te0jy(const regor::Operation *op, const Context &context);
void ErrorIfCheck_gpp861oen43y(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2nanft1ivm5fj(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1ga3gcg4zkrkv(const regor::Operation *op, const Context &context);
void ErrorIfCheck_36r4wpx3psd81(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1lrylbkd3w7ix(const regor::Operation *op, const Context &context);
void ErrorIfCheck_ojmgqziimenu(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3vqy81ueu5wjk(const regor::Operation *op, const Context &context);
void ErrorIfCheck_125xuezh1964i(const regor::Operation *op, const Context &context);
void ErrorIfCheck_fqta626ku4qe(const regor::Operation *op, const Context &context);
void ErrorIfCheck_ycjhrvf2yigr(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1c57olj698f3d(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1hrio849y2qnx(const regor::Operation *op, const Context &context);
void ErrorIfCheck_31vgfyg6fi9t6(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3fzsq78v5ypau(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2vhj6e48eyzlr(const regor::Operation *op, const Context &context);
void ErrorIfCheck_147wc580l2tik(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1gr4n0iszdlxr(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2rm8rnsdfn14h(const regor::Operation *op, const Context &context);
void ErrorIfCheck_36emtx7zwkk96(const regor::Operation *op, const Context &context);
void ErrorIfCheck_cr43yjpqkcpd(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3m5ijs493bw6j(const regor::Operation *op, const Context &context);
void ErrorIfCheck_341t6ysqc16b2(const regor::Operation *op, const Context &context);
void ErrorIfCheck_uqm570jwaqb6(const regor::Operation *op, const Context &context);
void ErrorIfCheck_34iiwt6o66qfa(const regor::Operation *op, const Context &context);
void ErrorIfCheck_llbd3iugmek0(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1w510kxt5b2b2(const regor::Operation *op, const Context &context);
void ErrorIfCheck_27g3t38z1of4h(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2cpco8ykx99sa(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2d0jmyhr9lscf(const regor::Operation *op, const Context &context);
void ErrorIfCheck_10td4qt70dp3i(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1qxtjwwlh068t(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1hp4djlq1mi8i(const regor::Operation *op, const Context &context);
void ErrorIfCheck_20r08ymi6c43u(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1xwwkxeypcw3j(const regor::Operation *op, const Context &context);
void ErrorIfCheck_vi3hzxbetjyg(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1m8qk2pbuovev(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1iv4j2x95j8dk(const regor::Operation *op, const Context &context);
void ErrorIfCheck_316kdwzc9jf5x(const regor::Operation *op, const Context &context);
void ErrorIfCheck_tnr115b4spgw(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2autvayhidla8(const regor::Operation *op, const Context &context);
void ErrorIfCheck_h1uadv5irsu6(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1kfh97qingywb(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1azcq4511qzyx(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2befn2dfjcm62(const regor::Operation *op, const Context &context);
void ErrorIfCheck_13tqdu59nyxyh(const regor::Operation *op, const Context &context);
void ErrorIfCheck_khc2s3en2uxi(const regor::Operation *op, const Context &context);
void ErrorIfCheck_q9dl3x81rc4o(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2rfkujt9lg7eq(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3nelbnmxyemot(const regor::Operation *op, const Context &context);
void ErrorIfCheck_24conlof4w8eh(const regor::Operation *op, const Context &context);
void ErrorIfCheck_xod9coigx1x2(const regor::Operation *op, const Context &context);
void ErrorIfCheck_15y4an3ceern5(const regor::Operation *op, const Context &context);
void ErrorIfCheck_10u6py7exa66n(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1hynqeiugz9lt(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1yism57if6v2z(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3k5ug2w7gxc7r(const regor::Operation *op, const Context &context);
void ErrorIfCheck_396rg8p65j58r(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3l2ksvk26m07h(const regor::Operation *op, const Context &context);
void ErrorIfCheck_192e2vu3t5aqm(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3tccsjner0km9(const regor::Operation *op, const Context &context);
void ErrorIfCheck_33exz9gn2i1wy(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2d3qdl1f70i6y(const regor::Operation *op, const Context &context);
void ErrorIfCheck_5y7ov1oeymoa(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1aloht2b77zby(const regor::Operation *op, const Context &context);
void ErrorIfCheck_f1kt9a6h7s2p(const regor::Operation *op, const Context &context);
void ErrorIfCheck_302z1f8mq8lg7(const regor::Operation *op, const Context &context);
void ErrorIfCheck_14z7y0qe9lwps(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3dvn5k3273lwz(const regor::Operation *op, const Context &context);
void ErrorIfCheck_34zvbtwx1r18j(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2a1jpygblc07i(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3hthyoock2ew5(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1nifeiq9rvmb8(const regor::Operation *op, const Context &context);
void ErrorIfCheck_21rq6kn6p1yle(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3rghkieqip43o(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1cyv9n59wyyyc(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3oy2tclc6uhsu(const regor::Operation *op, const Context &context);
void ErrorIfCheck_gpp3enlp1ddg(const regor::Operation *op, const Context &context);
void ErrorIfCheck_ix9div4ld46q(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3estuseky2gm2(const regor::Operation *op, const Context &context);
void ErrorIfCheck_5bq1fx1llv8(const regor::Operation *op, const Context &context);
void ErrorIfCheck_ckwpttzajw06(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2n1ratxgd89tx(const regor::Operation *op, const Context &context);
void ErrorIfCheck_aizwrn95lb0l(const regor::Operation *op, const Context &context);
void ErrorIfCheck_294afuxnedk9i(const regor::Operation *op, const Context &context);
void ErrorIfCheck_27p0n0pjt2bd6(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1uwmsen32dse1(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3c5bq3iswjd1x(const regor::Operation *op, const Context &context);
void ErrorIfCheck_53yuoon46swi(const regor::Operation *op, const Context &context);
void ErrorIfCheck_q9pgbwuvutqu(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1qdcccs22lqtr(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2azl8wc8mbsrj(const regor::Operation *op, const Context &context);
void ErrorIfCheck_122a36k26p0au(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3sfcy967j2w8w(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1obslcewwn583(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3oxfjen91qb6l(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1uo0z247e42af(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1eovh9pyc6tyw(const regor::Operation *op, const Context &context);
void ErrorIfCheck_24jsin2zkf4ug(const regor::Operation *op, const Context &context);
void ErrorIfCheck_12uj5fltk5rbo(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1py9f91imwjxe(const regor::Operation *op, const Context &context);
void ErrorIfCheck_fn614zzdrdfd(const regor::Operation *op, const Context &context);
void ErrorIfCheck_338aejy0aeqeg(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2a4sjfbd544h5(const regor::Operation *op, const Context &context);
void ErrorIfCheck_32ylwe00j5q2l(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3uwlzew8kfq5w(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1sxf726x838dv(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2fl3he9sci345(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1acxf2776vdap(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2ntycki2dof18(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1yv98jo1xcmke(const regor::Operation *op, const Context &context);
void ErrorIfCheck_bkdiivlz937z(const regor::Operation *op, const Context &context);
void ErrorIfCheck_242iuwska81dr(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2vooovn86b8fd(const regor::Operation *op, const Context &context);
void ErrorIfCheck_107z2k4den74o(const regor::Operation *op, const Context &context);
void ErrorIfCheck_38712gnuluf0u(const regor::Operation *op, const Context &context);
void ErrorIfCheck_4alci0dog4gp(const regor::Operation *op, const Context &context);
void ErrorIfCheck_31ty7f0kcbfxg(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3oet4aggtv528(const regor::Operation *op, const Context &context);
void ErrorIfCheck_15kl5g5u1jrhq(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1bm39avugkqqd(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3tv3oatlz37e2(const regor::Operation *op, const Context &context);
void ErrorIfCheck_n7biu53x2n6k(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2fd4dk1zw032u(const regor::Operation *op, const Context &context);
void ErrorIfCheck_omgw2xdm6irr(const regor::Operation *op, const Context &context);
void ErrorIfCheck_2jyu87hs8upt4(const regor::Operation *op, const Context &context);
void ErrorIfCheck_12uu5ff3t3lv8(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3puzf7van5acf(const regor::Operation *op, const Context &context);
void ErrorIfCheck_8tihij7a5ep0(const regor::Operation *op, const Context &context);
void ErrorIfCheck_3lu68v2531bjz(const regor::Operation *op, const Context &context);
void ErrorIfCheck_1fzl0zyxyd88z(const regor::Operation *op, const Context &context);
}  // namespace checks
}  // namespace validator
}  // namespace tosa
